#![feature(iterator_fold_self)]
use itertools::Itertools;
use rayon::prelude::*;
use std::collections::LinkedList;

fn count_and_say(s: &[u32]) -> Vec<(usize, u32)> {
    s.iter()
        .group_by(|e| *e)
        .into_iter()
        .map(|(key, group)| (group.into_iter().count(), *key))
        .collect()
}

fn par_count_and_say(s: &[u32]) -> Vec<(usize, u32)> {
    rayon::iter::split(s, |s| {
        if s.len() < 2 {
            (s, None)
        } else {
            let mid = s.len() / 2;
            let (left, right) = s.split_at(mid);
            (left, Some(right))
        }
    })
    .map(|s| count_and_say(s))
    .map(|v| std::iter::once(v).collect())
    .reduce(LinkedList::new, |mut l1, mut l2| {
        l1.append(&mut l2);
        l1
    })
    .into_iter()
    .fold_first(|mut v1, v2| {
        let skipped = v1
            .last_mut()
            .and_then(|(ref mut last_count, last_key)| {
                v2.first().map(|(first_count, first_key)| {
                    if first_key == last_key {
                        *last_count += first_count;
                        1
                    } else {
                        0
                    }
                })
            })
            .unwrap_or(0);
        v1.par_extend(v2.into_par_iter().skip(skipped));
        v1
    })
    .unwrap_or_else(Vec::new)
}

const SIZE: usize = 100_000_000;
fn main() {
    let input = [2, 2, 2, 3, 3, 4, 1, 1, 1];
    assert_eq!(count_and_say(&input), [(3, 2), (2, 3), (1, 4), (3, 1)]);

    let random_input = std::iter::repeat_with(|| rand::random::<u32>() % 10)
        .take(SIZE)
        .collect::<Vec<_>>();
    let start = std::time::Instant::now();
    let seq_count = count_and_say(&random_input);
    println!("seq: {:?}", start.elapsed());
    let start = std::time::Instant::now();
    let par_count = par_count_and_say(&random_input);
    println!("par: {:?}", start.elapsed());
    assert_eq!(seq_count, par_count);
}
